package ru.ulstu.tsMethods;

import ru.ulstu.TimeSeriesUtils;
import ru.ulstu.models.Forecast;
import ru.ulstu.models.Model;
import ru.ulstu.models.TimeSeries;
import ru.ulstu.models.TimeSeriesValue;
import ru.ulstu.models.exceptions.ForecastValidateException;
import ru.ulstu.models.exceptions.ModelingException;
import ru.ulstu.models.exceptions.TimeSeriesValidateException;

import java.time.temporal.ChronoUnit;

/**
 * Наиболее общая логика моделировани и прогнозирования временных рядов
 */
public abstract class TimeSeriesMethod {
    /**
     * Возвращает модельное представление временного ряда: для тех же точек времени что и в параметре timeSeries
     * строится модель. Количество точек может быть изменено: сокращено при сжатии ряда, увеличено при интерполяции.
     * Метод является шаблонным, выполняет операции валидации исходного ряда и потом его моделирование
     *
     * @param timeSeries исходный временной ряд подлежащий моделированию
     * @return модель временного ряда
     * @throws TimeSeriesValidateException
     */
    public Model getModel(TimeSeries timeSeries) throws TimeSeriesValidateException {
        validateTimeSeries(timeSeries);
        return getModelOfValidTimeSeries(timeSeries);
    }

    /**
     * Возвращает модельное представление валидного временного ряда: для тех же точек времени что и в параметре timeSeries
     * строится модель. Количество точек может быть изменено: сокращено при сжатии ряда, увеличено при интерполяции.
     *
     * @param timeSeries исходный временной ряд подлежащий моделированию
     * @return
     */
    protected abstract Model getModelOfValidTimeSeries(TimeSeries timeSeries);

    /**
     * Выполняет построение прогноза временного ряда. Даты спрогнозированных точек будут сгенерированы по модельным точкам.
     *
     * @param model       модель временного ряда
     * @param countPoints количество точек для прогнозирования
     * @return прогноз временного ряда
     */
    public Forecast getForecast(Model model, int countPoints) throws ModelingException {
        Forecast forecast = new Forecast(model);
        forecast = generateEmptyForecastPoints(forecast, countPoints);
        forecast = makeForecast(forecast);
        if (!forecast.getForecastTimeSeries().getFirstValue()
                .equals(forecast.getModel().getModelTimeSeries().getLastValue())) {
            throw new ForecastValidateException("Первая точка прогноза должна совпадать с последней модельной точкой");
        }
        return forecast;
    }

    /**
     * Выполняет построение прогноза для уже сгенерированных будущих точек временного ряда.
     *
     * @param forecast Заготовка прогноза временного ряда с пустыми значениями
     * @return
     */
    protected abstract Forecast makeForecast(Forecast forecast);

    protected Forecast generateEmptyForecastPoints(Forecast forecast, int countPointForecast) {
        long diffMilliseconds = TimeSeriesUtils.getTimeDifferenceInMilliseconds(forecast.getModel().getOriginalTimeSeries());
        forecast.getForecastTimeSeries()
                .addValue(new TimeSeriesValue(forecast.getModel().getModelTimeSeries().getLastValue().getDate()));
        for (int i = 1; i < countPointForecast + 1; i++) {
            forecast.getForecastTimeSeries()
                    .addValue(new TimeSeriesValue(forecast.getForecastTimeSeries().getValues().get(i - 1).getDate().plus(diffMilliseconds, ChronoUnit.MILLIS)));
        }
        return forecast;
    }

    /**
     * Выполняет построение модели и прогноза временного ряда. Даты спрогнозированных точек будут сгенерированы
     * по модельным точкам.
     *
     * @param timeSeries  временной ряда
     * @param countPoints количество точек для прогнозирования
     * @return прогноз временного ряда
     */
    public Forecast getForecast(TimeSeries timeSeries, int countPoints) throws ModelingException {
        validateForecastParams(countPoints);
        return getForecast(getModel(timeSeries), countPoints);
    }

    private void validateForecastParams(int countPoints) throws ForecastValidateException {
        if (countPoints < 1) {
            throw new ForecastValidateException("Количество прогнозных точек должно быть больше 0");
        }
    }

    private void validateTimeSeries(TimeSeries timeSeries) throws TimeSeriesValidateException {
        if (timeSeries == null || timeSeries.isEmpty()) {
            throw new TimeSeriesValidateException("Временной ряд должен быть не пустым");
        }
        if (timeSeries.getLength() < 2) {
            throw new TimeSeriesValidateException("Временной ряд должен содержать хотя бы 2 точки");
        }
        if (timeSeries.getValues().stream().anyMatch(val -> val == null || val.getValue() == null)) {
            throw new TimeSeriesValidateException("Временной ряд содержит пустые значения");
        }
        if (timeSeries.getValues().stream().anyMatch(val -> val.getDate() == null)) {
            throw new TimeSeriesValidateException("Временной ряд должен иметь отметки времени");
        }
    }
}
